# multiplexed replica protocol for milestoning

# Parameters for the milestoning colvars
set nr [numReplicas]
set replica_id [myReplica]

global replicanum; set replicanum 20
set EACH_REVERSALS 1000; #number of trajectories launched per milestone
set NUM_REVERSALS [expr $replicanum * $EACH_REVERSALS]; #total number of trajectories launched
global outputdir; set outputdir ./output
global BWD_file; set BWD_file "$outputdir/backward_output.txt"
global FWD_file; set FWD_file "$outputdir/forward_output.txt"
global RESTART_freq; set RESTART_freq 10
global DCD_freq; set DCD_freq 1000
global RUN_freq; set RUN_freq 10
set eval_freq 2; # number of timesteps between two updates; this number is smaller than $RESTART_freq
set selective_runs 1000
global selective_list; set selective_list ""

set input_dcd "./example/input_%s.dcd"
set outputname $outputdir/$replica_id/ms.$replica_id
coordinates ./example/dipeptide.pdb
temperature 300

source multiplexed_colvar.tcl
source milestoning.namd

replicaUniformPatchGrids yes

seed [expr int(0*srand(int(100000*rand()) + 100*$replica_id) + 100000*rand() + 1)]

colvars	on
cv config $config

#############################################################
# multiplexed algorithm
#############################################################

if { ! [myReplica] } {
  set work_queue {}
  set worker_stack {}
  proc workers_idle {} {
    global worker_stack
    return [llength $worker_stack]
  }
  proc push_worker {w} {
    global worker_stack
    lappend worker_stack $w
  }
  proc pop_worker {} {
    global worker_stack
    set w [lindex $worker_stack end]
    set worker_stack [lreplace $worker_stack end end]
    return $w
  }
  proc push_work {w} {
    global work_queue
    if { $w != {} } {
      lappend work_queue $w
    }
  }
  proc pop_work {} {
    global work_queue
    set w [lindex $work_queue 0]
    set work_queue [lreplace $work_queue 0 0]
    return $w
  }
  proc enqueue_work_0 work {
    if [workers_idle] {
      replicaSend $work [pop_worker]
    } else {
      push_work $work
    }
  }
  proc dequeue_work_0 worker {
    global work_queue
    if [llength $work_queue] {
      return [pop_work]
    }
    if { [workers_idle] == [numReplicas] - 1 } {
      while { [workers_idle] } {
        replicaSend break [pop_worker]
      }
      return break
    }
    push_worker $worker
    return {}
  }
}

proc enqueue_work work { ;# this is the function we call when we want to assign a job to a replica
  replicaEval 0 [list enqueue_work_0 $work]
}

proc schedule_work {} {
  while { 1 } {
    set w [replicaEval 0 "dequeue_work_0 [myReplica]"]
    if { $w == {} } { 
      set w [replicaRecv 0]
    }
    eval $w
  }
}

#############################################################
#ms_run controls the backward and forward trajectories
#############################################################

  set ABORT_ON_CROSSING False
  set flag_self False
  set MAX_NUM_STEPS 1000000
  set stepnum 0
  set incubation_start 0

proc ms_run {i checkpoint_name {start_stage "backward"} {start_on_step 0}} {
  print "ms_output: starting job: $i on replica: [myReplica]"
  global aborting
  global backward_failed
  global stepnum
  global incubation_start
  global STAGE ;# one of 'backward' or 'forward'
  global ABORT_ON_CROSSING ;# whether to halt the job immmediately once a milestone is crossed
  global flag_self ;# whether we count a transition to the source milestone: "True" for backward stage
  global ID
  global RUN_freq
  global ms_id ;# which cells the milestone is currently located
  global EACH_REVERSALS
  global replicanum
  global selective_list 
  global selective_runs
  global outputdir
  global MAX_NUM_STEPS
  set backward_failed "False"
  set aborting "False"
  set stepnum $start_on_step
  set incubation_start 0
  set STAGE "backward"
  set ABORT_ON_CROSSING "True"
  set flag_self "True"
  set ID $i
  set ms_id "0 0"
  set search_id 0
  for {set cell_search 0} {$cell_search<[expr $replicanum*2]} {incr cell_search} {
	if {[lsearch $selective_list $cell_search]>=0} {
		set temp_incr $selective_runs
	} else {
		set temp_incr $EACH_REVERSALS
	}
	set search_id [expr $search_id+$temp_incr]
        if {$search_id > $i} {
                break
        }
  }
  set now_cell $cell_search; # note this now_cell is current cell number multiplied by 2
  set now_real_cell [expr $now_cell/2]
  if {[expr $now_cell%2]==0} {
	set cell_nb [expr ($now_real_cell+1)%$replicanum]
  } else {
	set cell_nb [expr ($now_real_cell-1)%$replicanum]
  }
  lset ms_id 0 $now_real_cell
  lset ms_id 1 $cell_nb

  outputname $outputdir/backward.$i
  dcdfile $outputdir/backward.$i.dcd
  restartname $outputdir/backward.$i.restart
  checkpointLoad $checkpoint_name global

  if {$start_stage == "backward"} {
    rescalevels -1 ;# reverse all velocities from the checkpoint
    print "ms_output: Running backward stage job: $i, ms_id=([lindex $ms_id 0], [lindex $ms_id 1]) on replica: [myReplica]"
    run $RUN_freq ;# run the first N steps
    while {$aborting == "False"} { ;# the backward cycle
      run norepeat $RUN_freq ;# continue to run N steps until $aborting is set to "True"
    }
    set aborting "False"
    set stepnum 0 ;# reset the step number
    set incubation_start 0 ;# reset the starting step number
  }
  if {$backward_failed == "False" } { ;# we are running forwards
    print "ms_output: Running forward stage job: $i, ms_id=([lindex $ms_id 0], [lindex $ms_id 1]) on replica: [myReplica]"
    set STAGE forward
    set ABORT_ON_CROSSING True 
    set flag_self False
    outputname $outputdir/forward.$i
    dcdfile $outputdir/forward.$i.dcd ;# rename the various NAMD outputs
    restartname $outputdir/forward.$i.restart
    checkpointLoad $checkpoint_name global
    run $RUN_freq ;# run the first N steps
    while {$aborting == "False"} { ;# the forward run cycle
      run norepeat $RUN_freq ;# continue to run N steps until $aborting is set to "True"
    }
  }
}

startup
replicaBarrier

if { ! [myReplica] } { ;# procedure for launching jobs
  array set bwd_dict {} ;# create an empty backward index dictionary (array)
  array set fwd_dict {} ;# create an empty forward index dictionary (array)
  # Check for existence of completed backward and forward files
  if {[ file exists $BWD_file ]} {
    print "Reading backward completion file: $BWD_file"
    set bwdfile [open $BWD_file r] ;# open the backbward file
    set lines [split [read $bwdfile] "\n"]
    foreach line $lines { ;# each of the lines contains info about the jobs that were running, whether they finished or need to be restarted
      set index [lindex $line 0] ;# the index of the job
      set outcome [lindex $line 1] ;# whether the backward stage was a success
      if { $index != "" } {set bwd_dict($index) $outcome} ;# fill out the dictionary with this info
    }
  }
  
  if {[ file exists $FWD_file ]} {
    set fwdfile [open $FWD_file r] ;# open the forward file: this has been written by the milestoning.tcl script while the sims were running
    set lines [split [read $fwdfile] "\n"]
    foreach line $lines { ;# each of the lines contains info about the jobs that were running, whether they finished or need to be restarted
      #  a. place entry into an dictionary: all values will be 1
      set index [lindex $line 0] ;# the index of the job
      set outcome [lindex $line 1] ;# whether the forward stage was a success
      set fwd_dict($index) $outcome ;# they will all have a value of 1
    }
  }

  reinitvels $temperature
  set start_stage "backward"
#  set start_stage "forward"

  for { set i 0 } { $i < $NUM_REVERSALS } { incr i } { ;# for all possible backward we are going to run
    set search_id 0
    for {set cell_search 0} {$cell_search<[expr $replicanum*2]} {incr cell_search} {
        if {[lsearch $selective_list $cell_search]>=0} {
                set temp_incr $selective_runs
        } else {
                set temp_incr $EACH_REVERSALS
        }
        set search_id [expr $search_id+$temp_incr]
        if {$search_id > $i} {
                break
        }
    }
    set now_cell $cell_search

    if {[info exists bwd_dict(job$i)] && ($bwd_dict(job$i)=="failed")} {
	print "Job $i backward already completed and failed, skipping..."
	continue ;# move on to the next job
    }
    if {[info exists fwd_dict(job$i)] && ($fwd_dict(job$i)=="succeeded")} {
	print "Job $i forward already completed, skipping..."
	continue ;# move on to the next job
    }
    coorfile open dcd [format $input_dcd $now_cell]
    set frame_now [expr $temp_incr - ($search_id - $i)]

    for {set tmp_frame 0} { $tmp_frame < $frame_now } {incr tmp_frame} {
      coorfile skip
    }
    set result [coorfile read]
    print "coordinates set for job id $i, frame $frame_now (in cell $now_cell)"
    if {$result == -1} {
      coorfile close
      print "Requested backward runs exceeds DCD file length. Skipping all jobs after this point.";
      break
    }
###################################################################################################
#for restarting from an old run 
###################################################################################################
    set stepnum 0
    set incubation_start 0
    set start_stage "backward"
    if {[file exists "$outputdir/backward.$i.restart.coor"] && [file exists "$outputdir/backward.$i.restart.vel"] && [file exists "$outputdir/backward.$i.restart.xsc"] && [info exists bwd_dict(job$i)] } { ;# then commence from the restart files if not finished; otherwise just go with the dcd coords and random vels
      if {[file size "$outputdir/backward.$i.restart.coor"]==0||[file size "$outputdir/backward.$i.restart.vel"]==0||[file size "$outputdir/backward.$i.restart.xsc"]==0} {
      } elseif {$bwd_dict(job$i)!="succeeded"} {
      	print "Job $i backward may have not finished, restarting from the point where the restarts left off..."
      	reinitatoms $outputdir/backward.$i.restart
        set stepnum $bwd_dict(job$i) ;# restart the step number that we left off at
        set incubation_start 0
        print "Resetting stepnum: $stepnum, and incubation_start: $incubation_start"
      }
    }
    
    if {[ info exists fwd_dict(job$i) ]} { ;# then the forward never completed, it must be restarted
      if {$fwd_dict(job$i)=="succeeded"} {
	print "ERROR! failed initial forward-info reading check."
      }
      set start_stage "forward"
      if {[file exists "$outputdir/forward.$i.restart.coor"] && [file exists "$outputdir/forward.$i.restart.vel"] && [file exists "$outputdir/forward.$i.restart.xsc"] } { ;# then commence from the restart files
	if {[file size "$outputdir/forward.$i.restart.coor"]==0||[file size "$outputdir/forward.$i.restart.vel"]==0||[file size "$outputdir/forward.$i.restart.xsc"]==0} {
	} else {
		reinitatoms $outputdir/forward.$i.restart
		print "Job $i forward may have not finished, restarting from the point where the restarts left off..."
		set stepnum $fwd_dict(job$i) ;# restart the step number that we left off at
		set incubation_start 0
		print "Resetting stepnum: $stepnum, and incubation_start: $incubation_start"
	}
      }
    }
###################################################################################################    
###################################################################################################

    set checkpoint_name "system$i" ;# assign this system's checkpoint for easy loading by the assigned replica
    checkpointStore $checkpoint_name global ;# store this checkpoint to be reawakened later
    enqueue_work [list ms_run $i $checkpoint_name $start_stage $stepnum] ;# send this job to the stack to be pulled out by another replica
    reinitvels $temperature
    print "done initializing the systems for step $i"    
    coorfile close
  }
}

schedule_work



exit
